#include "../../internal-include/BootLoader/boot_loader.hpp"
#include "../../internal-include/GPU/gpu_internal.hpp"
#include <PSX/System/IOPorts/dma_io.hpp>
#include <stdio.h>

#include <PSX/GTE/gte.hpp>
#include <PSX/System/syscalls.hpp>
#include "../../reference/inline_n.h"
#include <math.h>

extern "C" uint32_t __heap_start;
extern "C" uint32_t __bss_start;
extern "C" uint32_t __bss_end;
extern "C" uint32_t __bss_len;
extern "C" uint32_t __planschi_start;
extern "C" uint32_t __planschi_end;

namespace JabyEngine {
    namespace boot {
        namespace Start {
            static void enable_DMA() {
                DMA_IO::DPCR.write(DMA_IO::DPCR.read().set(DMA_IO::DPCR::SPUEnable, DMA_IO::DPCR::GPUEnable, DMA_IO::DPCR::CDROMEnable));
            }

            static void setup() {
                static constexpr auto DebugX     = 1;
                static constexpr auto DebugY     = 0;
                static constexpr auto DebugScale = 1.0;

                __debug_boot_color_at(::JabyEngine::GPU::Color24::White(), DebugX, DebugY, DebugScale);
                Periphery::setup();

                __debug_boot_color_at(::JabyEngine::GPU::Color24::White(), DebugX, DebugY, DebugScale);
                enable_DMA();

                __debug_boot_color_at(::JabyEngine::GPU::Color24::Red(), DebugX, DebugY, DebugScale);
                SPU::stop_voices();
                
                __debug_boot_color_at(::JabyEngine::GPU::Color24::Green(), DebugX, DebugY, DebugScale);
                CD::setup();
                __debug_boot_color_at(::JabyEngine::GPU::Color24::Blue(), DebugX, DebugY, DebugScale);
                Timer::setup();

                __debug_boot_color_at(::JabyEngine::GPU::Color24::Yellow(), DebugX, DebugY, DebugScale);
                GPU::setup();
                GPU::display_logo();
                GTE::setup();
                
                //Pause??
                SPU::setup();
            }
        }
    }

    static void test_bios_font() {
        static constexpr uint16_t SJIS = 0x83B5;

        const auto*      font     = SysCall::Krom2RawAdd(SJIS);
        const auto*const font_end = font + 16;
        printf("Loading SJIS from @0x%p\n", font);
        
        for(; font < font_end; font++) {
            const auto cur_char = __builtin_bswap16(*font);
            for(uint16_t shift = 1 << 15; shift; shift = shift >> 1) {
                printf(cur_char & shift ? "#" : " ");
            }

            printf("\n");
        }
    }

    static void test_gte() {
#define gte_MulMatrix0(r1,r2,r3)				\
                {	gte_SetRotMatrix(r1);	\
                    gte_ldclmv(r2);		\
                    gte_rtir();		\
                    gte_stclmv(r3);		\
                    gte_ldclmv((char*)r2+2);\
                    gte_rtir();		\
                    gte_stclmv((char*)r3+2);\
                    gte_ldclmv((char*)r2+4);\
                    gte_rtir();		\
                    gte_stclmv((char*)r3+4);	}
                    
        const auto m0 = GTE::MATRIX::identity();
        const auto m1 = GTE::MATRIX::identity();
        auto       m2 = GTE::MATRIX::identity();

        asm("# PLANSCHI START");
        JabyEngine::GTE::multiply_matrix(m0, m1, m2);
        asm("# PLANSCHI END");

        #define PI 3.14159265 // v 90°???

        static const auto display_stuff = [](int32_t angle) {
            static const auto wanna_be_float = [](int32_t value) -> int32_t {
                return (value*10000)/(4096);
            };

            const auto sin_value = sin(GTE::in_degree(angle));
            const auto cos_value = cos(GTE::in_degree(angle));
            
            printf("Sine of %i = %i (%i)\n", angle, sin_value, wanna_be_float(sin_value));
            printf("Cos  of %i = %i (%i)\n", angle, cos_value, wanna_be_float(cos_value));
        };

        static const auto print_matrix = [](const GTE::MATRIX& matrix) {
            printf("{%i, %i, %i}\n", matrix.rot[0][0], matrix.rot[0][1], matrix.rot[0][2]);
            printf("{%i, %i, %i}\n", matrix.rot[1][0], matrix.rot[1][1], matrix.rot[1][2]);
            printf("{%i, %i, %i}\n", matrix.rot[2][0], matrix.rot[2][1], matrix.rot[2][2]);
            printf("{%i, %i, %i}\n", matrix.trans[0],  matrix.trans[1],  matrix.trans[2]);
        };

        display_stuff(25);
        display_stuff(45);
        display_stuff(75);
        display_stuff(90);

        auto rot_mat = []() -> GTE::MATRIX {
            auto rot = GTE::MATRIX::identity();

            rot.rot[0][0] = 123;
            rot.rot[0][1] = 456;
            rot.rot[0][2] = 789;

            rot.rot[1][0] = 1123;
            rot.rot[1][1] = 1456;
            rot.rot[1][2] = 1789;

            rot.rot[2][0] = 2123;
            rot.rot[2][1] = 2456;
            rot.rot[2][2] = 2789;

            rot.trans[0] = 9;
            rot.trans[1] = 8;
            rot.trans[2] = 7;
            return rot;
        }();

        print_matrix(rot_mat);
        printf("---\n");
        GTE::set_matrix(rot_mat);
        rot_mat = GTE::get_matrix();
        print_matrix(rot_mat);
        printf("---\n");
    }

    void start() {
        static constexpr auto DebugX     = 0;
        static constexpr auto DebugY     = 0;
        static constexpr auto DebugScale = 1.0;

        __debug_boot_print_at(GPU::Color24::White(), DebugX, DebugY, DebugScale, "Starting Planschbecken\n");
        printf("Heap starts @0x%p\n", &__heap_start);
        printf("BSS      from 0x%p to 0x%p (%u)\n", &__bss_start, &__bss_end, __bss_len);
        test_bios_font();
        test_gte();
        __debug_boot_print_at(GPU::Color24::Green(), DebugX, DebugY, DebugScale, "PLANSCHI from 0x%p to 0x%p\n", &__planschi_start, &__planschi_end);
        boot::Start::setup();
        
        __debug_boot_print_at(GPU::Color24::Red(), DebugX, DebugY, DebugScale, "Running main...\n");
        GPU::internal::wait_vsync(0);
        run();
    }
}